package calc;


import java.util.*;


/** Parses the user input and constructs Operational Tree*/

public class ExprParser {

  static class Debug {
    private static void out(String str) {
      //System.out.println(str);
    }
  }

  private StringTokenizer st;


  // whether a token has been consumed by expr, term ... etc.
  private boolean consumed = true;
  // the root node of the Operational Tree
  private BaseCalc root;

  // the current node of the Operational Tree
  private BaseCalc currentNode;

  // the registers for the current token values
  private double numtok;
  private String symtok;

  // the typecode describes the type of the current token
  private static final int TC_NONE = 0;
  private static final int TC_VAR = 1;
  private static final int TC_NUM = 2;
  private static final int TC_SYM = 3 ;
  private static final int TC_FUN = 4;
  private static final int TC_EOT = -1;
    
  private int typeCode;
    
  /** @param input the String to be parsed */
  public ExprParser(String input) throws SyntaxErrorException {
    String token;
    st = new StringTokenizer(input, "^+-*/()iIxXSsCcEe \t",true);

    // recursively construct Operational Tree
    root = expr();
  }
    
    
  /** returns the constructed Operational Tree by its root node */
  public BaseCalc getOperationTree() {
    return root;
  }

  /** returns a Vector with all nodes of type <code>VariableCalc</code>
      used for the insertion of values */
  public Vector getVariableCalcs(){
    Vector retVal = new Vector();
    if(root instanceof VariableCalc)
      retVal.addElement(root);
    walkChildren(retVal,root);
    return retVal;
  }
    

  /** recursively walks through the Operational tree, inserting instances
      of type <code>VariableCalc</code> into <code>parent</code> */
  private void walkChildren(Vector retVal, BaseCalc parent){
    BaseCalc children[] = parent.getChildren();
    if(children != null) 
      for(int i = 0; i < children.length; i++){
	if(children[i] instanceof VariableCalc)
	  retVal.addElement(children[i]);
	walkChildren(retVal, children[i]);
      }
  }
    
  /** the Lexer: produces Tokens and stores them (depending on type) in
      the appropriate global registers. Returns false if finished */
  private boolean getToken() throws SyntaxErrorException {
    String token;
    if(!consumed){
      Debug.out("Token has not been consumed");
      return true;
    }
    consumed = false;

    try {
      do 
	token = st.nextToken();
      while(token.equals(" ") || token.equals("\t"));
    } catch (NoSuchElementException e) { 
      return false; 
    }
	
    Debug.out("token is: '" + token+"'");


    // parse a numeral
    if(Character.isDigit(token.charAt(0))) {
      Debug.out("Parsing Numeral ...");
      int i;
      for(i = 0; i < token.length()
	    &&  (Character.isDigit(token.charAt(i)) 
		 || token.charAt(i) == '.'); i++);
	    
      if(i != token.length()) 
	throw new SyntaxErrorException();
      try {
	numtok = Double.valueOf(token).doubleValue();
      } catch (NumberFormatException E) {
	throw new SyntaxErrorException(); }
      typeCode = TC_NUM;
      return true;
    }

    // parse i = sqrt(-1)
    if(token.equals("i")){
      Debug.out("Parsing Numeral: i ...");
      numtok = Double.NaN;
      typeCode = TC_NUM;
      return true;
    }
      

	    
    // parse a variable
    if(token.charAt(0) == 'x' || token.charAt(0) == 't') {
      Debug.out("Parsing Variable ...");
      typeCode = TC_VAR;
      return true;
    }

    // parse a function
    if(token.equals("s") || token.equals("c")
       || token.equals("e") || token.equals("p") 
       || token.equals("m") || token.equals("r") 
       ||token.equals("")){
      Debug.out("Parsing Function ...");
      typeCode = TC_FUN;
      symtok = token;
      return true;
    }
	
    // must be an operation
    typeCode = TC_SYM;
    symtok = token;
    return true;
  }


  /** <pre>
      Implements the rule 
      <expr>	->	<term> { '+' <term> } | <term> { '-' <term> } . 
      </pre> */
  private BaseCalc expr() throws SyntaxErrorException {
	
    Debug.out("<expr>");

    BaseCalc retVal = null;
    BaseCalc firstTerm = term();

    //the arguments may be either term or the previous part of the expr
    BaseCalc arg1;
    BaseCalc arg2; 

    while(getToken() && typeCode == TC_SYM && 
	  (symtok.equals("+") || symtok.equals("-"))) {
	    
      consumed = true;
      BaseCalc operation;
	    
      // distinguish between addition and subtraction
      if(symtok.equals("+")) {
	operation = new AddCalc();
	Debug.out("Addition ...");
      }
      else {
	Debug.out("Subtraction ...");
	operation = new SubtractCalc();
      }

      // have we already performed an addition/subtraction in the 
      // while-loop? if yes, connect the subtree (stored in retVal) 
      // as a child to the new retVal
      if(retVal == null){
	arg1 = firstTerm;
	arg2 = term();
      }
      else {
	arg1 = retVal;
	arg2 = term();
      }

      retVal = operation;
	    
      // connect the children in the operational tree to the return Value
      BaseCalc[] children = {arg1,arg2};
      try {
	retVal.insertChildren(children);
      } catch (IllegalArgException e) { System.out.println(e);}
    }

    // if there is no addition or subtraction, simply return the first Term
    if(retVal == null)
      retVal = firstTerm;
    Debug.out("leaving <expr>");
    return retVal;
  }

  private BaseCalc term() throws SyntaxErrorException{
	
    Debug.out("<term>");
    BaseCalc retVal = null;
    BaseCalc factnode = fact();

    // Multiplication
    if(getToken() && typeCode == TC_SYM && symtok.equals("*")){
      consumed = true;
      Debug.out("Multiplication ...");

      retVal = new MultiplyCalc();
      BaseCalc factnode2 = fact();
	    
      BaseCalc[] children = {factnode, factnode2};
      try {
	retVal.insertChildren(children);
      } catch (IllegalArgException e) { System.out.println(e);}
    } else

      // Division
      if(typeCode == TC_SYM && symtok.equals("/")){
	consumed = true;
	Debug.out("Division ...");

	retVal = new DivCalc();
	BaseCalc factnode2 = fact();
	    
	BaseCalc[] children = {factnode, factnode2};
	try {
	  retVal.insertChildren(children);
	} catch (IllegalArgException e) { System.out.println(e);}
      } else
	retVal = factnode;

    if(retVal == null)
      throw new SyntaxErrorException();

    Debug.out("leaving <term>");
    return retVal;


  }


  private BaseCalc fact() throws SyntaxErrorException{
    BaseCalc retVal = null;
    Debug.out("<fact>");
    if(getToken() && typeCode == TC_FUN){
      consumed = true;

      // all the functions need only 1 argument and therefore 
      // differ only in Calculation (i.e. which BaseCalc to choose)
      if(symtok.equals("s"))
	retVal = new SinCalc();
      if(symtok.equals("c"))
	retVal = new CosCalc();
      if(symtok.equals("e"))
	retVal = new ExpCalc(); 
      if(symtok.equals(""))
	retVal = new ZetaCalc(); 
      if(symtok.equals("m"))
	retVal = new ImCalc(); 
      if(symtok.equals("r"))
	retVal = new RealCalc(); 
      BaseCalc arg = pot();
      BaseCalc children[] = {arg};	    
      try {
	retVal.insertChildren(children);
      } catch (IllegalArgException e) { System.out.println(e);}
	   
    } else
      retVal = pot();
	  
    Debug.out("leaving <fact>");
    return retVal;
  }

  private BaseCalc pot() throws SyntaxErrorException{
    Debug.out("<pot>");

    BaseCalc retVal = null;
    BaseCalc potnode = prim();

    // raising to power
    if(getToken() && typeCode == TC_SYM && symtok.equals("^")){
      consumed = true;
      Debug.out("Power ...");

      retVal = new PowerCalc();
      BaseCalc potnode2 = prim();
	    
      BaseCalc[] children = {potnode, potnode2};
      try {
	retVal.insertChildren(children);
      } catch (IllegalArgException e) { System.out.println(e);}
    } else 
      retVal = potnode;

    if(retVal == null)
      throw new SyntaxErrorException();
	
    Debug.out("leaving <pot>");
    return retVal;

  }

  private BaseCalc prim() throws SyntaxErrorException{
    BaseCalc retVal = null;
    Debug.out("<prim>");

	
    // check for an unary minus
    if(getToken() && typeCode == TC_SYM && symtok.equals("-")){
      Debug.out("setting minus ...");
      consumed = true;
      retVal = new MultiplyCalc();
      BaseCalc factnode = new NumeralCalc(new Complex(-1,0));
      BaseCalc factnode2 = fact();	      
      BaseCalc[] children = {factnode, factnode2};
      try {
	retVal.insertChildren(children);
      } catch (IllegalArgException e) { System.out.println(e);}
      Debug.out("leaving <prim>");
      return retVal;
    }
    

    // check for numeral
    if(typeCode == TC_NUM) {
      Debug.out("numtok is: "+numtok);      
      consumed = true;
      // check for i = sqrt(-1)
      if(Double.isNaN(numtok))
	retVal = new NumeralCalc(new Complex(0, 1));
      else
	retVal = new NumeralCalc(new Complex(numtok,0));      
      Debug.out("leaving <prim>");
      return retVal;
    }
    
    // check for variables
    if(typeCode == TC_VAR){
      Debug.out("Variable t");      
      consumed = true;
      retVal = new VariableCalc();
      Debug.out("leaving <prim>");
      return retVal;
    }

    if(typeCode == TC_SYM && symtok.equals("(")) {
      consumed = true;
      retVal = expr();
	  
      if(!getToken() || typeCode != TC_SYM || !symtok.equals(")")) 
	throw new SyntaxErrorException();
      else 
	consumed = true;
      Debug.out("leaving <prim>");
      return retVal;
    }
    
    System.out.println("leaving <prim> unexpectedly!!!!!");
    return retVal;

  }

  /** simply output all tokens */
  void test() {
    String token;
    try {
      while(st.hasMoreTokens()){
	token = st.nextToken();
	Debug.out(token);
      }
    } catch(Exception e){e.printStackTrace();}
  }

  /** For testing purposes */
  public static void main(String[] args) {
    try {
      ExprParser e = new ExprParser(args[0]);
      e.test();
      BaseCalc res = e.getOperationTree();
      System.out.println(res.toString());
      System.out.println("Result is: "+res.result());
    } catch (Exception e) { e.printStackTrace(); }
	
  }

}
